package LLI::Presence;

# This object should help us keep track of our data better, so we
# don't have to touch the data structures ourselves.
# 
## $Id: Presence.pm,v 1.1 2003/03/24 01:05:11 mankins Exp $

use strict;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(OFFLINE ONLINE UNKNOWN COMPONENT USER);

use constant UNKNOWN         =>   0;
use constant OFFLINE         =>   1;
use constant ONLINE          =>   2;

use constant COMPONENT       =>  10;
use constant USER            =>  20;

use constant REQUESTED       => 100;
use constant SUBSCRIBED      => 101;
use constant DENIED          => 102;
use constant UNREQUESTED     => 103;

our $VERSION = 1.00;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($jid) = (@_);
    
    my $self = {};
    bless ($self, $class);

    $self->add($jid) if ($jid);

    return $self;
  }

sub add
  {
    my $self = shift;

    my ($jid) = @_;

    # Initialize stuff.
    $self->jid($jid);
    $self->type();

    return $jid;
  }

sub jid
  {
    my $self = shift;

    my ($set) = @_;

    if ($set)
      {
	# We want to set the jid to this value.
	return $self->{'jid'} = $set;
      }
    else
      {
	return $self->{'jid'};
      }
  }

sub type
  {
    my $self = shift;

    my ($set) = @_;

    if ($self->{'jid'} =~ /\@/)
      {
	return USER;
      }
    else
      {
	return COMPONENT;
      }
  }

sub status
  {
    # this tells us if they are online or offline.  
    # It has nothing to do with subscriptions.

    my $self = shift;
    
    my %status;
    $status{"offline"} = OFFLINE;
    $status{"online"} = ONLINE;
    $status{"unknown"} = UNKNOWN;
    $status{OFFLINE} = "offline";
    $status{ONLINE} = "online";
    $status{UNKNOWN} = "unknown";

    my ($set) = @_;

    if ($set)
      {
	# We want to set the status to this value.
	
	my $stat = $status{$set};

	$stat = $stat || UNKNOWN;
	
	return $self->{'status'} = $stat;
      }
    else
      {
	return $status{$self->{'status'}}; # return words or constants?
      }
  }

sub subscription_status
  {
    # This tells us if we're subscribed to them, and if they're subscribed to us.
    #
    # It does this by returning a hash with keys "external" and "internal", meaning
    # is it a subscription to an external entity (we subscribe to them), or an
    # internal one (they subscribe to us).  The values for these keys should clue us in.

    my $self = shift;
    
    # 'requested','subscribed','denied','unrequested'

    my ($set) = @_;

    if (ref ($set) eq "HASH")
      {
	# We want to set the status to these values.

	my %subscriptions;
	$subscriptions{"external"} = $set->{"external"} if defined $set->{"external"};
	$subscriptions{"internal"} = $set->{"internal"} if defined $set->{"internal"};

	$self->{'subscriptions'} = \%subscriptions;
      }

    return $self->{'subscriptions'};
  }

sub is_subscriber
  {
    my $self = shift;
    
    my ($jid) = @_;

    return ($self->is_external_subscription($jid) && $self->is_internal_subscription($jid)) ? 1 : 0;
  }

sub requested_subscription
  {
    # this should help us boil this subscription nonsense down to a yes/no

    my $self = shift;

    my ($jid) = @_;

    return 1 if ($self->{'subscriptions'}->{"external"} eq 'requested');
    return 0;
  }

sub is_external_subscription
  {
    # this should help us boil this subscription nonsense down to a yes/no

    my $self = shift;

    my ($jid) = @_;

    return 1 if ($self->{'subscriptions'}->{"external"} eq 'subscribed');
    return 0;
  }

sub is_internal_subscription
  {
    # this should help us boil this subscription nonsense down to a yes/no

    my $self = shift;

    my ($jid) = @_;

    return 1 if ($self->{'subscriptions'}->{"internal"} eq 'yes');
    return 0;
  }

1;
