package LLI::RouteTable;

# This package manages/translates between the database store and Route objects
#
# $Id: RouteTable.pm,v 1.2 2003/03/28 04:28:12 mankins Exp $

my $VERSION = 1.00;

use strict;

#use Net::Jabber qw(Client);
use XML::Twig;
use Data::Dumper;

use LLI::Route;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($self) = (@_);

    $self->{'table'}             = $self->{'table'} || 'routes';

    bless ($self, $class);

    # syncronize our data with the database
    $self->sync_database();

    return $self;
  }

sub update_loopback
  {
    # this sets our own route information

    my $self = shift;
    
    my ($jid,$props) = @_;

    return undef unless ($jid && $props);

    my $route = new LLI::Route;
    
    $route->jid($jid);
    $route->preference(0);
    $route->centroid_lat($props->{'centroid_lat'});
    $route->centroid_lon($props->{'centroid_lon'});
    $route->boundary_top("lat",$props->{'bound_lat1'});
    $route->boundary_top("lon",$props->{'bound_lon1'});
    $route->boundary_bottom("lat",$props->{'bound_lat2'});
    $route->boundary_bottom("lon",$props->{'bound_lon2'});
    
    # Now record this in our database
    $self->set($route);

    return $route;
  }

sub add
  {
    # this is just a helper function that returns a minimally "valid" Route and
    # uses it to setup a stub in the database.  
    #
    # $routetable->add($jid);

    my $self = shift;

    my ($jid) = @_;

    return undef unless $jid;

    my $route = new LLI::Route;
    
    $route->jid($jid);
    $route->preference(0);
    
    # Now record this in our database
    $self->set($route);

    $self->get_route($jid);

    return $route;
  }

sub sync_database
  {
    my $self = shift;

    # we're calling this each time the database is change...
    # it should have been done so that the two are sync'd

    my $dbh = $self->{'dbh'};

    my $statement = "SELECT * FROM ".$self->{table};
    my $result = $dbh->selectall_hashref($statement, "routeid", undef );    

    my (@routes);

    # if we have anything in our database, populate
    if ($result)
      {

	# Go through our results from the database and fill out our memory structures
	foreach my $routeid (keys %{ $result })
	  {
	    # Create our LLI::Route object with the dat
	    my $route = new LLI::Route;
	    $route->jid($result->{$routeid}->{"jid"});
	    $route->preference($result->{$routeid}->{"preference"});
	    
	    $route->boundary_top("lat",$result->{$routeid}->{"bound_lat1"});
	    $route->boundary_bottom("lat",$result->{$routeid}->{"bound_lat2"});
	    $route->boundary_top("lon",$result->{$routeid}->{"bound_lon1"});
	    $route->boundary_bottom("lon",$result->{$routeid}->{"bound_lon2"});
	    
	    $route->centroid_lat($result->{$routeid}->{"centroid_lat"});
	    $route->centroid_lon($result->{$routeid}->{"centroid_lon"});
	    
	    push (@routes, $route);
	  }
      }

    return $self->{'routes'} = \@routes;
  }

sub jid_exists
  {
    my $self = shift;

    my ($jid) = @_;

    my $dbh = $self->{'dbh'};

    my $statement = "SELECT routeid FROM ".$self->{table}." WHERE jid = ?";
    my $result = $dbh->selectrow_hashref($statement, undef, ($jid) );

    return ($result->{'routeid'});
  }

sub delete
  {
    my $self = shift;

    my ($jid) = @_;

    my $dbh = $self->{'dbh'};

    my $statement = "DELETE FROM ".$self->{table}." WHERE jid = ?";
    my $result_rows = $dbh->do($statement, undef, ($jid) );

    $self->sync_database();

    return $result_rows;
  }

sub get_route
  {
    my $self = shift;
    my ($jid) = @_;

    # rename stuff for no good reason...
    my $c = $self->{jabber_connection};
    my $dbh = $self->{'dbh'};

    return if (lc $jid eq lc $self->{'jabber_from'});
    
    # Send request through our Jabber Connection and wait for an answer
    my $xml_request = "<iq type='get' xmlns='lli:routes' id='".time."' from='".$self->{'jabber_from'}."' to='".$jid."'><query xmlns='lli:routes' jid='$jid'/></iq>";
    &LLI::debug("xml $xml_request");

    my $result = $c->Send($xml_request);
    return unless $result;
   
    my $twig = XML::Twig->new(twig_roots   => { 'route' => 1} );
    $twig->safe_parse($result->GetXML());
    my (@routes) = $twig->get_xpath("route") if $twig;

    foreach my $r (@routes)
      {
	my $jid = $r->att("jid");
	my (@group) = $r->children("group");

	foreach my $g (@group)
	  {
	    # get ahold of the data
	    my $preference = $g->att("preference");
	    my ($boundary) = $g->find_nodes("boundary");
	    my ($centroid) = $g->find_nodes("centroid");

	    my ($bottom) = $boundary->descendants("bottom");
	    my ($top) = $boundary->descendants("top");
	    my ($bottom_lat) = $bottom->att("lat");
	    my ($top_lat) = $top->att("lat");
	    my ($bottom_lon) = $bottom->att("lon");
	    my ($top_lon) = $top->att("lon");

	    my ($centroid_lat) = $centroid->att("lat");
	    my ($centroid_lon) = $centroid->att("lon");

	    # Create the route object and fill in the pieces
	    my $route = new LLI::Route;

	    $route->jid($jid);
	    $route->preference($preference);
	    $route->centroid_lat($centroid_lat);
	    $route->centroid_lon($centroid_lon);
	    $route->boundary_top("lat",$top_lat);
	    $route->boundary_top("lon",$top_lon);
	    $route->boundary_bottom("lat",$bottom_lat);
	    $route->boundary_bottom("lon",$bottom_lon);

	    # Now record this in our database
	    $self->set($route);
	  }
      }
  }

sub prime_routes
  {
    my $self = shift;

    # The idea here is that if we've never connected to the network before
    # or somehow have an empty route table we need to be able to figure out
    # where the network is and how we route
    #
    # So we ask a known good host if they can help us out...

    my ($wait_time) = @_; # time to wait until we give up getting help from this known good

    $wait_time = $wait_time || 30;

    # rename stuff for no good reason...
    my $c = $self->{jabber_connection};
    my $dbh = $self->{'dbh'};
    
    # Send request through our Jabber Connection and wait for an answer
    my $xml_request = "<iq type='get' xmlns='lli:routes' from='".$self->{'jabber_from'}."' to='".$self->{'primer_host'}."'><query xmlns='lli:routes'/></iq>";
    my $result = $c->SendAndReceiveWithID($xml_request,$wait_time);
    return unless $result;

    my $twig = XML::Twig->new(twig_roots   => { 'route' => 1} );
    $twig->safe_parse($result->GetXML());
    my (@routes) = $twig->get_xpath("route") if $twig;

    foreach my $r (@routes)
      {
	my $jid = $r->att("jid");
	my (@group) = $r->children("group");

	foreach my $g (@group)
	  {
	    # get ahold of the data
	    my $preference = $g->att("preference");
	    my ($boundary) = $g->find_nodes("boundary");
	    my ($centroid) = $g->find_nodes("centroid");

	    my ($bottom) = $boundary->descendants("bottom");
	    my ($top) = $boundary->descendants("top");
	    my ($bottom_lat) = $bottom->att("lat");
	    my ($top_lat) = $top->att("lat");
	    my ($bottom_lon) = $bottom->att("lon");
	    my ($top_lon) = $top->att("lon");

	    my ($centroid_lat) = $centroid->att("lat");
	    my ($centroid_lon) = $centroid->att("lon");

	    # Create the route object and fill in the pieces
	    my $route = new LLI::Route;

	    $route->jid($jid);
	    $route->preference($preference);
	    $route->centroid_lat($centroid_lat);
	    $route->centroid_lon($centroid_lon);
	    $route->boundary_top("lat",$top_lat);
	    $route->boundary_top("lon",$top_lon);
	    $route->boundary_bottom("lat",$bottom_lat);
	    $route->boundary_bottom("lon",$bottom_lon);

	    # Now record this in our database
	    $self->set($route);
	  }
      }
  }

sub set
  {
    # Takes an array of route objects and sets that in the database

    # my $route = {
    #            'route'  => { 'jid' => 'lli.mit.edu',
    #			      'group' => 
    #			      [
    #			       { 'preference' => '0',
    #				 'centroid' =>
    #				 {
    #				  'lat' => ['42'],
    #				  'lon' => ['-71']
    #				 },
    #				 'boundary' =>
    #				 {
    #				  'top' => 
    #				  {
    #				   'lat' => ['42'],
    #				   'lon' => ['-71']
    #				  },
    #				  'bottom' =>
    #				  {
    #				   'lat' => ['42'],
    #				   'lon' => ['-71']
    #				  }
    #				 }
    #			       },
    #			      ]
    #			    };

    my $self = shift;

    my $dbh = $self->{'dbh'};

    my ($routes) = @_;

    # Convert to an array ref if just one
    unless (ref($routes) eq 'ARRAY')
      {
	$routes = [ $routes ];
      }

    foreach my $route (@{$routes})
      {

      LOOK_FOR_EXISTING_ENTRY:
	{
	  my $statement = "SELECT routeid FROM ".$self->{table}." WHERE jid = ? AND preference = ?";
	  my $result = $dbh->selectrow_hashref($statement, undef, ($route->jid(),$route->preference()) );    	

	  my $matching_row = $result->{"routeid"};

	  if ($matching_row)
	    {
	      # Found a match
	    UPDATE_DATABASE:
	      {
		my (%p,@vals);
		$p{'bound_lat1'} = $route->boundary_top("lat");
		$p{'bound_lat2'} = $route->boundary_bottom("lat");
		$p{'bound_lon1'} = $route->boundary_top("lon");
		$p{'bound_lon2'} = $route->boundary_bottom("lon");
		$p{'centroid_lat'} = $route->centroid_lat();
		$p{'centroid_lon'} = $route->centroid_lon();
		
		my (@params) = ("bound_lat1","bound_lat2","bound_lon1","bound_lon2","centroid_lat","centroid_lon");
		my $statement = "UPDATE ".$self->{table}." SET ";

		foreach my $param (@params)
		  {
		    if ($p{$param})
		      {
			# zero problem exists here.
			$statement .= "$param = ?,";
			push (@vals,$p{$param});
		      }
		  }

		chop($statement); # remote last comma
		$statement .= " WHERE routeid = ?";
		push(@vals,$matching_row);
		
		next if (scalar(@vals) == 1); # no need to update if null query

		my $result = $dbh->do($statement, undef, (@vals));
		
		&LLI::debug("did $statement",@vals); 
		
		# should check return value and do something?
	      }
	    }
	  else
	    {
	      # No matches found, let's add it

	    ADD_NEW_ROUTE_ROW:
	      {
		my $statement = "INSERT INTO ".$self->{table}." SET bound_lat1 = ?, bound_lat2 = ?, bound_lon1 = ?, bound_lon2 = ?, centroid_lat = ?, centroid_lon = ?, jid = ?, preference = ?";
		my $result = $dbh->do($statement, undef, ($route->boundary_top("lat"), $route->boundary_bottom("lat"), $route->boundary_top("lon"), $route->boundary_bottom("lon"), $route->centroid_lat(), $route->centroid_lon(), $route->jid(), $route->preference()));

		# should check return value and do something?
	      }
	    }
	}
      }

    $self->sync_database();
  }

sub get
  {
    my $self = shift;

    my $dbh = $self->{'dbh'};

    my ($jid) = @_;

    return unless $self->{"routes"};

    my (@matched_routes);

    foreach my $route (@{ $self->{"routes"} })
      {
	# if this is one of the ones we want, save it.
	if ($route->jid() eq $jid)
	  {
	    push (@matched_routes,$route);
	  }
      }

    return \@matched_routes;
  }

sub clear
  {
    # used to delete a jid's rows from the route table.  If called without an argument, clears 
    # whole table.  use with caution...

    my $self = shift;

    my ($jid) = @_;

    my $dbh = $self->{'dbh'};

    my ($statement,$result);

    if ($jid)
      {
	$statement = "DELETE FROM ".$self->{table}." WHERE jid = ?";
	$result = $dbh->do($statement, undef, ($jid) );    
      }
    else
      {
	$statement = "DELETE FROM ".$self->{table};
	$result = $dbh->do($statement, undef, () );    
      }

    # remember these changes
    $self->sync_database();

    return $result;
  }

sub size
  {
    # Look to see how big our route table is

    my $self = shift;

    my $dbh = $self->{'dbh'};

    my $statement = "SELECT COUNT(DISTINCT jid) FROM ".$self->{table};
    my ($result) = $dbh->selectrow_array($statement);    

    return ($result);
  }


sub get_jids
  {
    # Get all of the JID's from route table

    my $self = shift;

    my $dbh = $self->{'dbh'};

    my $statement = "SELECT DISTINCT jid FROM ".$self->{table};
    my ($result) = $dbh->selectall_arrayref($statement);    

    print Dumper $result;
 
    my (@results);

    #$VAR1 = [
    #          [
    #            'lli.jabber.semicolons.com'
    #          ],
    #          [
    #            'lli.mit.edu'
    #          ]
    #        ];

    foreach my $r (@{$result})
      {
	push(@results,@{$r}) if (ref $r);
      }

    return (\@results);
  }

sub get_routes_search
{
  my $self = shift;

  my ($position,$radius) = @_;

  my $dbh = $self->{'dbh'};

  $position->units_latlon("decimal");
  my $lat = $position->get_lat();
  my $lon = $position->get_lon();

  &LLI::Searcher::debug("lat $lat lon $lon ".$position->units_latlon());

  my ($lat1,$lat2,$lon1,$lon2) = ( ($lat + $radius), ($lat - $radius), ($lon + $radius), ($lon - $radius) );

  my $statement = "SELECT jid FROM ".$self->{table}." WHERE bound_lat1 <= $lat2 AND bound_lon1 >= $lon1 AND bound_lat2 >= $lat1 AND bound_lon2 <= $lon2 ORDER BY preference";

  &LLI::Searcher::debug($statement);
  &LLI::Searcher::debug(($lat1,$lat2,$lon1,$lon2));

  my $result = $dbh->selectall_arrayref($statement);
  return $result;
}

1;

