#!/usr/bin/perl

# This is mainly used for testing...and is provided for reference.
# It's not really supposed to "do" anything...
#
# Simple Jabber Bot - Mr. Matt Mankins <mankins@media.mit.edu>
#
# $Id: test.cgi,v 1.1 2003/05/13 20:40:53 mankins Exp $

use strict;

use Net::Jabber qw(Client);
use XML::Simple;
use Data::Dumper;
use Digest::SHA1 qw(sha1_hex);

# Some constants
use constant NS_REGISTER => 'jabber:iq:register';

my $debug = 1;
sub debug;

$| = 1; # unbuffered

# Globals
my ($id,$from);
my $cfg = XMLin("lli-test-client.xml") || die "Can't load config file: $!\n";
my $component_name = "lli.jabber.media.mit.edu";
my ($jid,$resource);

print "Content-type: text/html\n\n";
debug "<PRE>";

my $c = new Net::Jabber::Client( debuglevel => $cfg->{'debug'}->{'level'}, debugfile => $cfg->{'debug'}->{'file'});

$c->Connect(hostname=> $cfg->{'client'}->{'hostname'}, port=> $cfg->{'client'}->{'port'}, ssl=> $cfg->{'client'}->{'ssl'});

if ($c->Connected()) 
  {
    debug "We are connected to the server...\n". $cfg->{'client'}->{'hostname'}."\n";
    
  AUTHENTICATE:
    {
      # Try to authenticate
      my ($status,$error) = &authenticate();

      # Review the status
      if ($status !~ /^ok/i)
	{
	  print "status: $status\n";
	  if ($cfg->{'client'}->{'autocreate'})
	    {
	      debug "Automatically creating the user account.";
	      my $status2 = create_account({ 'username' => $cfg->{'client'}->{'username'},
			  'password' => $cfg->{'client'}->{'password'}});
	
	      if ($status2)
		{
		  print "Account created.\n";
		}
	      else
		{
		  print "Unable to create account.\n";
		}
	    }
	  else
	    {
	      # We should return the error to the client
	      print "Username/Password incorrect. Auto-create not set.";
	    }

          $c->Disconnect;
	  exit; # Try again
	}

      $from = $cfg->{'client'}->{'username'}."\@".$cfg->{'client'}->{'hostname'}."\/".$cfg->{'client'}->{'resource'};
      ($jid,$resource) = split('/',$from);

      debug "connected and authenticated as $from";
    }

  SETUP_CALLBACKS:
    {
      debug "setting up callbacks";

      $c->SetPresenceCallBacks(     available       =>      \&presence_avail,
				    unavailable     =>      \&presence_unavail,
			      );

    }

  SEND_PRESENCE:
    {
      next; # commented out for now.

      # Sending my presence
      debug "sending presence";

      my $Presence = Net::Jabber::Presence->new();
      $Presence->SetType("available");
      $Presence->SetStatus("Online");
      $c->Send($Presence);

      $c->PresenceSend("Online");
    }
    
  CONNECT_TO_COMPONENT:
    {
      #start comment out various sections to test...

    PUBLISH_TO_NODE_CELL:
      {
	print "publishing to cell node...<BR>";

	my ($content) = get_test_content();

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".web_escape_xml($result->GetType)." ".web_escape_xml($result->GetErrorCode())."<BR>\n" if $result;
	print web_escape_xml($result->GetXML()) if $result;
	print "<BR><BR>-- end of publish to cell --<BR><BR>";
      }

    PUBLISH_TO_NODE_PRESENCE:
      {
	print "publishing to position node...<BR>";

	my ($content) = get_test_content_position("p:$jid","62.11","-71.5","3",undef,"current");

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml(($result->GetXML())) if $result;
	print "\n<BR>--end publish<BR><BR>";
      }


  DO_ASYNC_QUERY:
    {
      #      last;

      my $node_id;
    SEND_SEARCH_FORM:
      {
	print "\nfilling out search form...\n";

	my ($result) = fill_search_form( { 'component' => "$component_name" } );

	next unless $result;

        print "\nfill form result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n";
	print web_escape_xml($result->GetXML())."\n";
      }
      
      last;

    CREATE_INSTA_NODE:
      {
        print "creating insta node...<BR>";
	
        # Try to create our node.  Returns result type: result, error, etc, and error code (409)
        my ($result) = create_node( { 'component' => "$component_name" } );

	next unless $result;

        print "create node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n";
	print web_escape_xml($result->GetXML())."\n";

	# This is one way of doing it, but maybe it's not general? (??)
	# my ($xml) = get_xml_object($result->GetXML());

	# So we'll do it this way where we walk the tree until we find what we want.

	($node_id) = get_data_from_node($result, 
					 [ 
					  { 
					   "name" => 'pubsub',
					   "attrib" => 'xmlns',
					   "value" => 'http://jabber.org/protocol/pubsub'
					  },
					  {
					   "name" => 'create',
					   "attrib" => 'node'
					  }
					 ]
					);


	print "found the node: $node_id\n";
      }

    REQUEST_SEARCH_FORM:
      {
	print "getting search form...\n";

	my ($result) = get_search_form( { 'component' => "$component_name" } );

	next unless $result;

        print "\nrequest search result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n";
	print web_escape_xml($result->GetXML())."\n";
      }

    SEND_SEARCH_FORM:
      {
	print "\nfilling out search form...\n";

	my ($result) = fill_search_form( { 'component' => "$component_name" } );

	next unless $result;

        print "\nfill form result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n";
	print web_escape_xml($result->GetXML())."\n";
      }

    }

    SUBSCRIBE_TO_PRESENCE_NODE:
      {
	print "subscribing to node...<BR>";

	foreach my $number (1,3,5,7,9)
	  {

	    my $jid = "tester$number\@jabber.media.mit.edu";
	    
	    my ($result) = subscribe_node("$component_name","p:$jid");

	    print "subscribe node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	    
	    print web_escape_xml($result->GetXML()) if $result;
	  }


	    print "<BR>--end subscribe<BR><BR>";

      }


    GET_ITEMS_FROM_NODE:
      {
	print "getting from node...<BR>";

	my $item = "current"; #"current";

	my ($result) = get_items_from_node("$component_name","p:tester7\@jabber.media.mit.edu",$item);
	print web_escape_xml($result->GetXML()) if $result;

	print "get node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print "\n<BR>--end get_items<BR><BR>";
      }

    SUBSCRIBE_TO_PRESENCE_NODE:
      {
	print "subscribing to node...<BR>";

	my $jid = "tester7\@jabber.media.mit.edu";

	my ($result) = subscribe_node("$component_name","p:$jid");

	print "subscribe node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml($result->GetXML()) if $result;
	print "<BR>--end subscribe<BR><BR>";
      }

    PUBLISH_TO_NODE_PRESENCE:
      {
	print "publishing to position node...<BR>";

	my ($content) = get_test_content_position("p:$jid","62.11","-71.5","3",undef,"current");

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml(($result->GetXML())) if $result;
	print "\n<BR>--end publish<BR><BR>";
      }

    GET_PENDING_SUBSCRIPTIONS:
      {
	print "getting pending subscriptions...<BR>";

	my ($result) = get_pending_subscriptions("$from","$component_name");

	print "publish node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml(($result->GetXML())) if $result;
	print "\n<BR>--end get pending<BR><BR>";
      }

    RETRACT_ITEM_FROM_NODE:
      {
	print "retract items from node...<BR>";
	
	my $item = "1047000473";

	my ($result) = retract_item("$component_name","p:$jid",$item);

	print "retract node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml($result->GetXML()) if $result;	
	print "\n<BR>---end retract---<BR>\n";
      }

    APPROVE_SUBSCRIPTION:
      {
	# this approves a subscription made previously.
	print "approving subscription...<BR>";

	my $node = "p:$jid";
	my $subscriber = "mankins\@jabber.media.mit.edu";

	my ($result) = approve_subscription("$node","$subscriber",1); # 1 = approved
	print web_escape_xml(($result->GetXML())) if $result;
        print "<BR>--end approve<BR><BR>";
      }

    GET_ITEMS_FROM_NODE:
      {
	print "getting from node...<BR>";

	my $item = ""; #"current";

	my ($result) = get_items_from_node("$component_name","p:$jid",$item);
	print web_escape_xml($result->GetXML()) if $result;

	print "get node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print "\n<BR>--end get_items<BR><BR>";
      }

    PUBLISH_TO_NODE_PRESENCE:
      {
	print "publishing to position node...<BR>";

	my ($content) = get_test_content_position("p:$jid","62.11","-71.5","3",undef,"myitem");

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml(($result->GetXML())) if $result;
	print "\n<BR>--end publish<BR><BR>";
      }

    PURGE_FROM_NODE:
      {
	print "purging items from node...<BR>";
	
	my $item = "current";

	my ($result) = purge_node("$component_name","p:asdf$from",$item);

	print "purge node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml($result->GetXML()) if $result;	
      }

    CREATE_NODE_PRESENCE:
      {
	print "creating node...<BR>";
	
	# Try to create our node.  Returns result type: result, error, etc, and error code (409)
	my ($result) = create_node( { "component" => "$component_name",
				      "node" => "p:$from" } );
	print "create node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml($result->GetXML()) if $result;
      }

    PUBLISH_TO_NODE_PRESENCE:
      {
	print "publishing to position node...<BR>";

	my ($content) = get_test_content_position("p:$from","52.11","-71.5","23",undef);

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print web_escape_xml(($result->GetXML())) if $result;
	print "<BR>--end publish<BR><BR>";
      }

    GET_ITEMS_FROM_NODE:
      {
	print "getting from node...<BR>";

	my $item = ""; #"current";

	my ($result) = get_items_from_node("$component_name","p:$jid",$item);
	print web_escape_xml($result->GetXML()) if $result;

	print "get node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
	print "<BR>--end get_items<BR><BR>";
      }

    DELETE_NODE:
      {
	print "deleting node...<BR>";

	my ($result) = delete_node("$component_name","p:$from");
	print "delete node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml(($result->GetXML())) if $result;
        print "<BR>--end delete<BR><BR>";
      }

    APPROVE_SUBSCRIPTION:
      {
	# this approves a subscription made previously.
	print "approving subscription...<BR>";

	my $node = "p:$jid";
	my $subscriber = "mankins\@jabber.media.mit.edu";

	my ($result) = approve_subscription("$node","$subscriber",1); # 1 = approved
	print web_escape_xml(($result->GetXML())) if $result;
        print "<BR>--end approve<BR><BR>";
      }

    SUBSCRIBE_TO_PRESENCE_NODE:
      {
	print "subscribing to node...<BR>";

	my ($result) = subscribe_node("$component_name","p:$jid");

	print "subscribe node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml($result->GetXML()) if $result;
	print "<BR>--end subscribe<BR><BR>";
      }

    PUBLISH_TO_NODE_CELL:
      {
	print "publishing to cell node...<BR>";

	my ($content) = get_test_content();

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".web_escape_xml($result->GetType)." ".web_escape_xml($result->GetErrorCode())."<BR>\n" if $result;
	print web_escape_xml($result->GetXML()) if $result;
	print "<BR><BR>-- end of publish to cell --<BR><BR>";
      }

    SUBSCRIBE_TO_CELL_NODE:
      {
	print "subscribing to node...<BR>";

	my ($result) = subscribe_node("$component_name","c:30,-70:1,1:14,07:04,24");

	print "subscribe node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml($result->GetXML()) if $result;
      }

    CREATE_NODE_CELL:
      {
	print "creating node...<BR>";
	
	# Try to create our node.  Returns result type: result, error, etc, and error code (409)
	my ($result) = create_node( { "component" => "$component_name",
				      "node" => "c:30,-70:1,1:14,07:04,24" } );
	print "create node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml($result->GetXML()) if $result;

	# cell nodes can't be created like this.
      }

    UNSUBSCRIBE_TO_NODE:
      {
	print "unsubscribing to node...<BR>";

	my ($result) = unsubscribe_node("$component_name","p:$jid");

	print "unsubscribe node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;

	print web_escape_xml($result->GetXML()) if $result;
      }

    DISAPPROVE_SUBSCRIPTION:
      {
	# this approves a subscription made previously.
	print "deny subscription...<BR>";

	my $node = "p:$jid";
	my $subscriber = "mankins\@jabber.media.mit.edu";

	my ($result) = approve_subscription("$node","$subscriber",0); # 0 = cancel
	print web_escape_xml(($result->GetXML())) if $result;
        print "<BR>--end deny<BR><BR>";
      }

    PUBLISH_TO_NODE_CELL:
      {
	print "publishing to cell node...<BR>";

	my ($content) = get_test_content();

	my ($result) = publish_node("$component_name","$from",$content);

	print "publish node result: ".web_escape_xml($result->GetType)." ".web_escape_xml($result->GetErrorCode())."<BR>\n" if $result;
	print web_escape_xml($result->GetXML()) if $result;
      }

  SAY_HELLO:
    {
      $c->MessageSend(to=>"mankins\@jabber.media.mit.edu",
		      subject=>"Hi there...",
		      body=>"$$\n",
		      thread=>"ABC123",
		      priority=>10);
    }

      
    GET_ITEMS_FROM_NODE_TWO:
      {
	print "getting from node...<BR>";
	
	my $item = "current";

	my ($result) = get_items_from_node("$component_name","$from",$item);

	print "get node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
      }

    DELETE_NODE:
      {
	print "deleting node...<BR>";

	my ($result) = delete_node("$component_name","$from");
	print "delete node result: ".$result->GetType." ".$result->GetErrorCode()."<BR>\n" if $result;
      }
      
    }

        
  DISCONNECT:
    {
      $c->Disconnect();
    }
    
  }

sub approve_subscription
  {
    #    <message from="node-owner" to="pubsub.jabber.org">
    #    <x xmlns="jabber:x:data" type="submit">
    #        <field var="FORM_TYPE">
    #           <value>http://jabber.org/protocol/pubsub</value>
    #        </field>
    #        <field var="node">
    #           <value>generic/pgm-mp3-player</value>
    #        </field>
    #        <field var="subscriber_jid">
    #           <value>sub1@foo.com</value>
    #        </field>
    #    </x>
    # </message>

    my ($node_name,$subscriber,$yes_or_no) = @_;

    if ($yes_or_no)
      {
	$yes_or_no = "submit"; # ok to subscribe
      }
    else
      {
	$yes_or_no = "cancel"; # deny the request
      }

    my $message = new Net::Jabber::Message;
    $message->SetFrom("$jid"); #node owner
    $message->SetTo($component_name); #component
    $message->SetThread(rand);

    my $x = $message->NewX("jabber:x:data");
    $x->SetType($yes_or_no);

    my $f = $x->AddField();
    $f->SetVar("FORM_TYPE");
    $f->SetValue("http://jabber.org/protocol/pubsub");

    $f = $x->AddField();
    $f->SetVar("node");
    $f->SetValue($node_name);

    $f = $x->AddField();
    $f->SetVar("subscriber_jid");
    $f->SetValue($subscriber);

    my $response = $c->Send($message); # subscription message to the node owner sent.

    print "message: <BR><BR>".web_escape_xml($message->GetXML())."<BR><BR>";

    return $response;
  }

sub get_children_from_node
  {
    my ($node) = @_;

    # Umm Ok, maybe this is a little hacky, but ...
    # It takes either a Net::Jabber Object, or the actual
    # XML::Stream::Node object and returns the children (as an XML::Stream::Node)

    my (@children);
    
    my ($tree);

    if (ref ($node) =~ 'Net::Jabber')
      {
	$tree = $node->GetTree();
      }
    else
      {
	$tree = $node;
      }

    if ($tree)
      {
	(@children) = $tree->children();
      }
    return (@children);
  }

sub get_data_from_node
  {
    my ($node, $data) = @_;

    # This takes either an XML::Stream::Node or a Net::Jabber object and looks
    # for a specific part of the tree to return.
    # 
    # Pass into $data a reference to an array, where each array is a hashref that
    # defines what you're looking for in this part of the tree
    #
    # We call this recursively to find our exact match, or undef

    # my ($thing) = get_data_from_node($result,
    #                                         [
    #                                          {
    #                                           "name" => 'pubsub',
    #                                           "attrib" => 'xmlns',
    #                                           "value" => 'http://jabber.org/protocol/pubsub'
    #                                          },
    #                                          {
    #                                           "name" => 'create',
    #                                           "attrib" => 'node'
    #                                          }
    #                                         ]
    #                                        );
    #
    # This should return  pubsub->create->node's value

    return ($node) unless ($node && $data); # if we don't know what to do, just return the tree

    my (@data) = @{ $data };

    my ($look_for_this) = shift(@data); # the current part of the tree we're after

    $data = \@data;

    my (@children) = get_children_from_node($node);

    my ($found);

    foreach my $child (@children)
	  {
	    if ($child->get_tag() eq $look_for_this->{"name"})
	      {
		next if ( ($look_for_this->{"attrib"} && defined($look_for_this->{"value"}) ) && ($child->get_attrib($look_for_this->{"attrib"}) !~ /$look_for_this->{"value"}/i) ); #case insensitive is okay?

		# Must have found what we're looking for
		$found = $child;
		last;
	      }
	  }

    unless (scalar(@children))
      {
	# If we have no children, then we found ourself
	$found = $node;
      }

    # Recurse if we have more to find.
    if ( ($data) && (scalar(@$data)) )
      {
	# We have more parts of the tree to look at.
	return get_data_from_node($found,$data);
      }
    else
      {
	# no more data, return what we found.
	if ($look_for_this->{"attrib"})
	  {
	    # if we're looking for something specific, return it
	    return $found->get_attrib($look_for_this->{"attrib"});
	  }
	else
	  {
	    # Otherwise, return the whole node
	    return $found;
	  }
      }

  }

sub get_xml_object
  {
    # Given an XML String, return an XML::Simple'd perl object that we can manipulate

    my ($xml_string) = @_;

    my ($xml) = eval
      {
	my $xs = new XML::Simple (forcearray => 1, keeproot => 1, keyattr => []);
	return $xs->XMLin($xml_string);
      };

    return $xml;
  }

sub presence_avail
  {
 #   debug "avail: ".join(':',@_)."\n";
  }

sub presence_unavail
  {
#    debug "unavail: ".join(':',@_)."\n";
  }

sub handle_incoming
  {
  #  debug "found: ".join(':',@_)."\n";
  }

sub handle_message
  {
#    debug "<BR>incoming message!";
#    debug "found: ".join(':',@_)."\n";
    my ($sid, $iq) = @_;

    print "incoming message!<BR>\n";
    print web_escape_xml($sid->GetXML());

  }

sub handle_iq
  {
    my ($sid, $iq) = @_;
    my @iqchildren = $iq->GetTree()->children();
    foreach my $iqc (@iqchildren)
      {
        if ($iqc->get_tag() eq "pubsub" && $iqc->get_attrib("xmlns") eq "http://jabber.org/protocol/pubsub")
          {
            my @pschildren = $iqc->children();
            foreach my $psc (@pschildren)
              {
                my $tag = $psc->get_tag();
		
                debug ("pubsub type:".$iq->GetType().":".$tag);
		
                if ($tag eq "configure" && $iq->GetType() eq "get")
                  {
                    pubsub_get_configure($iq, $psc);
                  }
                elsif ($tag eq "configure" && $iq->GetType() eq "set")
                  {
                    pubsub_set_configure($iq, $psc);
                  }
                elsif ($tag eq "create" && $iq->GetType() eq "set")
                  {
                    pubsub_set_create($iq, $psc);
                  }
                elsif ($tag eq "delete" && $iq->GetType() eq "set")
                  {
		     pubsub_set_delete($iq, $psc);
		   }
                elsif ($tag eq "purge" && $iq->GetType() eq "set")
                  {
                    pubsub_set_purge($iq, $psc);
                  }
                elsif ($tag eq "entities" && $iq->GetType() eq "get")
                  {
                    pubsub_get_entities($iq, $psc);
                  }
                elsif ($tag eq "entities" && $iq->GetType() eq "set")
                  {
                    pubsub_set_entities($iq, $psc);
                  }
                elsif ($tag eq "publish" && $iq->GetType() eq "set")
                  {
                    pubsub_set_publish($iq, $psc);
                  }
                elsif ($tag eq "subscribe" && $iq->GetType() eq "set")
                  {
                    pubsub_set_subscribe($iq, $psc);
                  }
                elsif ($tag eq "unsubscribe" && $iq->GetType() eq "set")
                  {
                    pubsub_set_unsubscribe($iq, $psc);
                  }
                elsif ($tag eq "items" && $iq->GetType() eq "get")
                  {
                    pubsub_get_items($iq, $psc);
		  }
		else
		  {
                    next;
                  }
                last;
              }
            last;
          }
        elsif ($iqc->get_tag() eq "query" && $iqc->get_attrib("xmlns") eq "http://jabber.org/protocol/disco#items")
          {
            pubsub_get_query($iq, $iqc);
            last;
          }
      }
  }

sub send_presence
  {

  }

sub create_xml
  {
    my ($object) = @_;

    # The problem: How do we convert a generic data structure into 
    #              XML?

    # The answer:  This sub.

    #$object = { 
    #	       'iq'  => { 'type' => 'set',
    #                     'from' => 'matt@moo.com',
    #			  'to'   => 'moo@matt.com',
    #			  'id'   => '1',
    #			  'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
    #					'create' => { 'node' => 'generic/foo' 
    #						    }
    #				      }
    #			} 
    #	      };

    # <opt>
    #  <iq id="1" to="moo@matt.com" from="matt@moo.com" type="set">
    #    <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #      <create node="generic/foo" />
    #    </pubsub>
    #  </iq>
    # </opt>
 
#    debug "incoming object:". Dumper $object;

    my ($xml) = eval 
                   {
		     my $xs = new XML::Simple (forcearray => 1, keeproot => 1, keyattr => []);
		     return $xs->XMLout($object);
		   };

    # Get rid of new lines (this is legal, right?)
    $xml =~ s/\>\n+\</\>\</g; 

    debug "<CODE>xml: ".web_escape_xml($xml)." . $@</CODE>";

    return $xml;
  }

sub web_escape_xml
  {
    my ($xml) = @_;

    # Make the XML readable in an HTML browser (as HTML)

    $xml =~ s/\>\</\>\n\</g; #add cr's

    $xml =~ s/\>/&gt;/g;
    $xml =~ s/\</&lt;/g;



    return $xml;
  }

sub get_test_content
  {
    #<iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="publish1">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <publish node="generic/pgm-mp3-player">
    #      <item id="current">
    #        <mp3 xmlns="mp3-ns">
    #           <song>Goldberg Variations</song>
    #           <artist>J.S. Bach</artist>
    #           <album>Going for Baroque</album>
    #        </mp3>
    #      </item>
    #    </publish>
    #  </pubsub>
    #</iq>

    # media lab is at:
    # Lat:  42.360624  42:21:38.246N 
    # Lon:  -071.088443  71:05:18.395W 

    my ($lat,$lon,$alt,$expiry,$type,$source,$url) = ();

    #    42:21:38 (lat) -71:05:18 (lon)
    #      lat: 43.360556 lon: -71.088333

    # c:10Dlat,10Dlon:1Dlat,1Dlon:Mlat,Mlon:Slat,Slon
    # c:-10,150:2,9:50,1:23,59 (cell)

    my $content = {
		   'node' => "c:40,-70:2,1:21,5:38,18",
		   'item' => [ 
			      {
			       'id' => 'current',

				'position' => 
				{ 
				 'xmlns' => 'lli-position',
				 
				  'lat' => ['42.360624'],
				  'lon' => ['-71.088443'],
				  'alt' => ['22'],
				},
			       'time' => 
			       {
				'xmlns' => 'lli-time',
				'timestamp' => ['1045604403'],
				'lifetime' => ['1055604403']
			       },
			       'link' => 
			       {
				'xmlns' => 'lli-link',
				'url' => ['http://jabber.media.mit.edu/~mankins/'],
				'cost' => ['0']
			       }
			      },
			     ]
		  };

    return $content;
  }

sub get_test_content_position
  {
    #<iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="publish1">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <publish node="generic/pgm-mp3-player">
    #      <item id="current">
    #        <mp3 xmlns="mp3-ns">
    #           <song>Goldberg Variations</song>
    #           <artist>J.S. Bach</artist>
    #           <album>Going for Baroque</album>
    #        </mp3>
    #      </item>
    #    </publish>
    #  </pubsub>
    #</iq>

    # media lab is at:
    # Lat:  42.360624  42:21:38.246N 
    # Lon:  -071.088443  71:05:18.395W 

    my ($node,$lat,$lon,$alt,$expiry,$type,$source,$url,$item_name) = (@_);

    my $item_name ||= "current";

    #    42:21:38 (lat) -71:05:18 (lon)
    #      lat: 43.360556 lon: -71.088333

    # c:10Dlat,10Dlon:1Dlat,1Dlon:Mlat,Mlon:Slat,Slon
    # c:-10,150:2,9:50,1:23,59 (cell)

    my $content = {
		   'node' => $node,
		   'item' => [ 
			      {
			       'id' => $item_name,

				'position' => 
				{ 
				 'xmlns' => 'lli-position',
				 
				  'lat' => [$lat],
				  'lon' => [$lon],
				  'alt' => [$alt],
				},
			       'time' => 
			       {
				'xmlns' => 'lli-time',
				'timestamp' => [time],
				'expiration' => [$expiry]
			       }
			      },
			     ]
		  };

    return $content;
  }

sub get_id
  {
    # this calculates the nugget ID for this nugget.

    my $self = shift;

    $self->{'id'} = sha1_hex($self->lat(),$self->lon(),$self->expiration(),$self->type(),$self->source(),$self->url());

    return $self->{'id'};
  }

sub retract_item
  {
    my ($host_name, $node_name, $item_name) = @_;

    # <iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="deleteitem1">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <retract node="generic/pgm-mp3-player">
    #      <item id="current"/>
    #    </retract>
    #  </pubsub>
    # </iq>

    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'retract' => { 
							 'node' => $node_name,
							 'item' => [
								    {
								     'id' => $item_name
								    }
                                                                 ]
							}
					 }
			   }		  
		 };
    
    my ($xml) = create_xml($pubsub);
    
    my ($response) = $c->SendAndReceiveWithID($xml,15);
    
    return $response;

  }

sub purge_node
  {
    my ($host_name, $node_name) = @_;
    
    # This removes all items from the node...
    
    #    <iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="purge1">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <purge node="generic/pgm-mp3-player"/>
    #  </pubsub>
    # </iq>
    
    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'purge' => { 'node' => $node_name }
					 }
			   }		  
		 };
    
    my ($xml) = create_xml($pubsub);
    
    my ($response) = $c->SendAndReceiveWithID($xml,15);
    
    return $response;
  }

sub get_pending_subscriptions
  {
    my ($from,$host_name) = @_;

    #<message from="node-owner" to="pubsub.jabber.org">
    #    <x xmlns="http://jabber.org/protocol/pubsub#owner"
    #       action="get-pending"/>
    #</message>

    my $xml = '<message from="'.$from.'" to="'.$host_name.'">
    <x xmlns="http://jabber.org/protocol/pubsub#owner"
       action="get-pending"/>
</message>';

    $c->Send($xml);
    $c->Process(10);
  }

sub subscribe_node
  {
    my ($host_name, $node_name) = @_;

    # <iq type="set" from="sub1@foo.com/home" to="pubsub.jabber.org" id="sub1">
    # <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #   <subscribe node="generic/pgm-mp3-player" jid="sub1@foo.com"/>
    # </pubsub>
    # </iq>

    my ($jid) = split_jid($from); # retieve just the user@hostname part of the JID
    
    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'subscribe' => { 'node' => $node_name,
							    'jid'  => $jid
						       }
					 }
			   }		  
		 };

    my ($xml) = create_xml($pubsub);

    print web_escape_xml($xml)."\n\n<BR>";

    my ($response) = $c->SendAndReceiveWithID($xml,15);

    return $response;
  }

sub split_jid
  {
    my ($jid_base) = @_;

    my $resource;
    ($jid_base,$resource) = split('/',$jid_base);

    return ($jid_base,$resource);
  }

sub unsubscribe_node
  {
    my ($host_name, $node_name) = @_;
    
    # <iq type="set" from="sub1@foo.com" to="pubsub.jabber.org" id="unsub1">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <unsubscribe node="generic/pgm-mp3-player"/>
    # </pubsub>
    # </iq>

    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'unsubscribe' => { 'node' => $node_name
						       }
					 }
			   }		  
		 };

    my ($xml) = create_xml($pubsub);

    my ($response) = $c->SendAndReceiveWithID($xml,15);

    return $response;
  }

sub publish_node
  {
    my ($host_name, $node_name, $content) = @_;

    # <iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="create1">
    #     <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #         <create node="generic/pgm-mp3-player"/> 
    #     </pubsub>
    # </iq>

    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'publish' => $content
					 }
			   }		  
		 };

    my ($xml) = create_xml($pubsub);

    print web_escape_xml($xml)."\n\n<BR>";

    my ($response) = $c->SendAndReceiveWithID($xml,15);

    return $response;
  }

sub get_items_from_node
  {
    my ($host_name, $node_name, $item) = @_;

    #<iq type="get" from="pgm@jabber.org" to="pubsub.jabber.org" id="items3">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <items node="generic/pgm-mp3-player">
    #      <item id="current"/>
    #    </items>
    #  </pubsub>
    #</iq>

    my $pubsub = {
		  'iq'  => { 'type' => 'get',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'items' => { 
						       'node' => $node_name,
						       'item' => [ 
								  {
								   'id' => $item
								  }
								 ]
						      }
					 }
			   }		  
		 };

    my ($xml) = create_xml($pubsub);

    my ($result) = $c->SendAndReceiveWithID($xml,15);

    return $result;        
  }

sub fill_search_form
  {
    my ($params) = @_;
    
    my $form_data = [
		     {
		      'var' => 'latitude',
		      'value' =>  [ '73' ],
		     },
		     {
		      'var' => 'longitude',
		      'value' =>  [ '44' ],
		     },
		     {
		      'var' => 'expiration',
		      'value' =>  [ '10' ],
		     },
		    ];

    $params->{"form_data"} = $form_data;

    return send_search_form($params);
  }

sub send_search_form
  {
    my ($params) = @_;
    
    # <iq type='set' from='bob@jabber.org' id="data3">
    #  <query xmlns='jabber:iq:search'>
    #   <x xmlns='jabber:x:data' type='submit'>
    #     <field var='gender'><value>male</value></field>
    #     <field var='first'><value>Bob</value></field>
    #   </x>
    #  </query>
    #  </iq>  

    my ($component) = $params->{"component"};
    my ($xmlns)     = $params->{"xmlns"} || 'jabber:iq:search';
    my ($my_from)   = $params->{"from"} || $from;

    my ($form_data) = $params->{"form_data"};

    my $iq = {
		  'iq'  => { 'type' => 'set',
			     'from' => $my_from,
			     'to'   => $component,
			     'query'=> {
					'xmlns' => $xmlns,
					'x' => 
					{
					 'xmlns' => 'jabber:x:data',
					 'type'  => 'submit',
					 'field' => $form_data,
					}
				       }
			   }		  
	     };
    
    my ($xml) = create_xml($iq);
    
    my ($response) = $c->SendAndReceiveWithID($xml,15);
    
    return $response;
  }

sub get_search_form
  {
    my ($params) = @_;

    # <iq type='get' to='bloodbank.jabber.org' id='data2'>
    #   <query xmlns='jabber:iq:search'/>
    # </iq>  

    my ($component) = $params->{"component"};
    my ($xmlns)     = $params->{"xmlns"} || 'jabber:iq:search';
    my ($my_from)   = $params->{"from"} || $from;

    my $iq = {
		  'iq'  => { 'type' => 'get',
			     'from' => $my_from,
			     'to'   => $component,
			     'query'=> { 'xmlns' => $xmlns }
			   }		  
		 };

    my ($xml) = create_xml($iq);

    my ($response) = $c->SendAndReceiveWithID($xml,15);

    return $response;
  }

sub create_node
  {
    my ($params) = @_;

    # <iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="create1">
    #     <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #         <create node="generic/pgm-mp3-player"/> 
    #     </pubsub>
    # </iq>

    my ($component) = $params->{"component"};
    my ($node_name) = $params->{"node"};

    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $component,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'create' => { 'node' => $node_name
						       }
					 }
			   }		  
		 };

    my ($xml) = create_xml($pubsub);

    my ($response) = $c->SendAndReceiveWithID($xml,15);

    return $response;
  }

sub delete_node
  {
    my ($host_name, $node_name) = @_;

    # <iq type="set" from="pgm@jabber.org" to="pubsub.jabber.org" id="delete1">
    #  <pubsub xmlns="http://jabber.org/protocol/pubsub">
    #    <delete node="generic/pgm-mp3-player"/>
    #  </pubsub>
    # </iq>

    my $pubsub = {
		  'iq'  => { 'type' => 'set',
			     'from' => $from,
			     'to'   => $host_name,
			     'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
					   'delete' => { 'node' => $node_name
						       }
					 }
			   }		  
		 };

    my ($xml) = create_xml($pubsub);

    my ($response) = $c->SendAndReceiveWithID($xml,15);

    return $response;
  }


sub get_next_id
  {
    return $id++;
  }

sub create_account
  {
    my ($details) = @_;

    #   <iq type='set'>
    #     <query xmlns='jabber:iq:register'>
    #       <username>...</username>
    #       <password>...</password>
    #     </query>
    #   </iq>

    my $status;

    my $iq = Net::Jabber::IQ->new();
    $iq->SetType('set');
    
    my $query = $iq->NewQuery(NS_REGISTER);

    debug "keys:".join(',',keys %{$details});

    foreach my $thing (keys %{$details})
      {
	debug "$thing = ".$details->{$thing};
      }

     $query->SetUsername($details->{"username"});
     $query->SetPassword($details->{"password"});

    # Could be more, I suppose.
      
    my $result = $c->SendAndReceiveWithID($iq);

    # Success
    if ($result->GetType() eq 'result') 
    {
       debug "registration ok";
       $status = 1;
    }
    else 
    {
       debug $result->GetErrorCode()." (",$result->GetError().")";
       $status = 0;
    }

    return $status;
  }


sub authenticate
  {
    print "trying to authenticcate\n\n" . $cfg->{'client'}->{'username'}. ":". $cfg->{'client'}->{'password'}."<BR>";
    my (@result) = $c->AuthSend('username' => $cfg->{'client'}->{'username'}, 'password' => $cfg->{'client'}->{'password'}, 'resource' => $cfg->{'client'}->{'resource'});

    return @result;
  }

sub debug
  {
    print join(':',@_,"\n") if $debug;
  }
